/*
 * libhfs - library for reading and writing Macintosh HFS volumes
 * Copyright (C) 1996-1998 Robert Leslie
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: amiga_fssm.c,v 1.2 2009/08/11 21:40:05 march Exp $
 */

# ifdef HAVE_CONFIG_H
#  include "config.h"
# endif

# include <errno.h>
# include <sys/stat.h>

# include "libhfs.h"
# include "os.h"

#include <dos/dos.h>
#include <dos/dosextens.h>
#include <dos/dostags.h>
#include <dos/filehandler.h>
#include <devices/trackdisk.h>

#include <proto/dos.h>
#include <proto/exec.h>

struct DeviceIO {
	struct FileSysStartupMsg *fssm;
	struct MsgPort *diskport;
	struct IOExtTD *ioreq;

	ULONG read_cmd;
	ULONG write_cmd;

	UQUAD bytes_start;
	UQUAD bytes_end;
};

static LONG InitDeviceIO(struct FileSysStartupMsg *fssm, struct DeviceIO *dio)
{
	APTR task;
	ULONG flags, unit;
	UBYTE *device;
	LONG err = 0;

	memset(dio, 0, sizeof(struct DeviceIO));

	dio->fssm = fssm;

	if (TypeOfMem(dio->fssm) == 0 || TypeOfMem(BADDR(dio->fssm->fssm_Device)) == 0 || TypeOfMem(BADDR(dio->fssm->fssm_Environ)) == 0)
		return ERROR_OBJECT_WRONG_TYPE;

	unit = dio->fssm->fssm_Unit;
	flags = dio->fssm->fssm_Flags;
	device = 1 + BADDR(dio->fssm->fssm_Device);

	if ((dio->diskport = CreateMsgPort()))
	{
		if ((dio->ioreq = CreateIORequest(dio->diskport, sizeof(struct IOExtTD))))
		{
			if (OpenDevice(device, unit, (struct IORequest *)dio->ioreq, flags) == 0)
			{
				struct DosEnvec *de = BADDR(dio->fssm->fssm_Environ);
				ULONG bytes_low, bytes_high;

				dio->bytes_start = 1LL * de->de_Surfaces * de->de_BlocksPerTrack * de->de_LowCyl * de->de_SizeBlock * 4;
				dio->bytes_end = 1LL * de->de_Surfaces * de->de_BlocksPerTrack * (de->de_HighCyl+1) * de->de_SizeBlock * 4;

				bytes_low = dio->bytes_start & 0xFFFFFFFF;
				bytes_high = dio->bytes_start >> 32;

				if (bytes_high)
				{
					dio->read_cmd = TD_READ64;
					dio->write_cmd = TD_WRITE64;
				}
				else
				{
					dio->read_cmd = CMD_READ;
					dio->write_cmd = CMD_WRITE;
				}

				return 0;
			}
			DeleteIORequest(dio->ioreq);
			dio->ioreq = NULL;
		}
		DeleteMsgPort(dio->diskport);
		dio->diskport = NULL;
	}
	dio->fssm = NULL;

	return ERROR_NO_FREE_STORE;
}

static LONG CleanupDeviceIO(struct DeviceIO *dio)
{
	if (dio->ioreq)
	{
		CloseDevice((struct IORequest *)dio->ioreq);
		DeleteIORequest(dio->ioreq);
	}

	if (dio->diskport)
		DeleteMsgPort(dio->diskport);

	dio->ioreq = NULL;
	dio->diskport = NULL;

	dio->fssm = NULL;
	return 0;
}

static LONG DoRawRead (struct DeviceIO *dio, ULONG n, void *buff, ULONG blocksize, ULONG count)
{
	UQUAD bytes_start;
	ULONG bytes_low, bytes_high;
	LONG err;
	bytes_start = dio->bytes_start + 1LL * n * blocksize;

	if (bytes_start >= dio->bytes_end || bytes_start < dio->bytes_start)
		return ERROR_SEEK_ERROR;

	bytes_low = bytes_start & 0xFFFFFFFF;
	bytes_high = bytes_start >> 32;

	dio->ioreq->iotd_Req.io_Command = dio->read_cmd;
	dio->ioreq->iotd_Req.io_Data = buff;
	dio->ioreq->iotd_Req.io_Length = blocksize*count;
	dio->ioreq->iotd_Req.io_Flags = 0;
	dio->ioreq->iotd_Req.io_Offset = bytes_low;
	dio->ioreq->iotd_Req.io_Actual = bytes_high;

	return DoIO((struct IORequest*)dio->ioreq);
}

static LONG DoRawWrite (struct DeviceIO *dio, ULONG n, void *buff, ULONG blocksize, ULONG count)
{
	UQUAD bytes_start;
	ULONG bytes_low, bytes_high;

	bytes_start = dio->bytes_start + 1LL * n * blocksize;

	if (bytes_start >= dio->bytes_end || bytes_start < dio->bytes_start)
		return ERROR_SEEK_ERROR;

	bytes_low = bytes_start & 0xFFFFFFFF;
	bytes_high = bytes_start >> 32;

	dio->ioreq->iotd_Req.io_Command = dio->write_cmd;
	dio->ioreq->iotd_Req.io_Data = buff;
	dio->ioreq->iotd_Req.io_Length = blocksize * count;
	dio->ioreq->iotd_Req.io_Flags = 0;
	dio->ioreq->iotd_Req.io_Offset = bytes_low;
	dio->ioreq->iotd_Req.io_Actual = bytes_high;

	return DoIO((struct IORequest*)dio->ioreq);
}

static ULONG GetTotalBlocks(struct DeviceIO *dio, int blocksize)
{
	UQUAD size;
	struct DosEnvec *de = BADDR(dio->fssm->fssm_Environ);

	size = 1LL * de->de_Surfaces * de->de_BlocksPerTrack * (de->de_HighCyl - de->de_LowCyl + 1) * (de->de_SizeBlock * 4) / blocksize;
	return size;
}

struct DeviceIO global_dio;
ULONG global_block_current;
ULONG global_total_blocks;

/*
 * NAME:	os->open()
 * DESCRIPTION:	open and lock a new descriptor from the given path and mode
 */
int os_open(void **priv, const char *path, int mode)
{
	LONG err;
	struct DeviceIO *dio = &global_dio;

	if ((err = InitDeviceIO((void *)path, dio)) == 0)
	{
		*priv = (void *)dio;
		global_total_blocks = GetTotalBlocks(dio, HFS_BLOCKSZ);
		global_block_current = 0;
		return 0;
	}
	ERROR(-1, "error opening medium");
fail:
	return -1;
}

/*
 * NAME:	os->close()
 * DESCRIPTION:	close an open descriptor
 */
int os_close(void **priv)
{
	struct DeviceIO *dio = *priv;
	CleanupDeviceIO(dio);
	*priv = NULL;
	return 0;
}

/*
 * NAME:	os->same()
 * DESCRIPTION:	return 1 iff path is same as the open descriptor
 */
int os_same(void **priv, const char *path)
{
	return 1; /* multiple devices are not supported yet */
}

/*
 * NAME:	os->seek()
 * DESCRIPTION:	set a descriptor's seek pointer (offset in blocks)
 */
unsigned long os_seek(void **priv, unsigned long offset)
{
	struct DeviceIO *dio = *priv;

	/* offset == -1 special; seek to last block of device */
	if (offset == (unsigned long) -1)
	{
		global_block_current = global_total_blocks;
//		  kprintf("os seek: block %ld, res %ld\n", offset, global_block_current);
		return global_block_current;
	}
	else if (offset >= 0 && offset < global_total_blocks)
	{
		global_block_current = offset;
//		  kprintf("os seek: block %ld, res %ld\n", offset, global_block_current);
		return global_block_current;
	}

//	  kprintf("os seek: block %ld, error\n", offset);
	ERROR(-1, "error seeking medium");
fail:
	return -1;
}

/*
 * NAME:	os->read()
 * DESCRIPTION:	read blocks from an open descriptor
 */
unsigned long os_read(void **priv, void *buf, unsigned long len)
{
	struct DeviceIO *dio = *priv;
	LONG err;

//	  kprintf("os read: count %ld\n", len);

	if ((err = DoRawRead(dio, global_block_current, buf, HFS_BLOCKSZ, len)) == 0)
		return len;

	ERROR(-1, "error reading from medium");
fail:
	return -1;
}

/*
 * NAME:	os->write()
 * DESCRIPTION:	write blocks to an open descriptor
 */
unsigned long os_write(void **priv, const void *buf, unsigned long len)
{
	struct DeviceIO *dio = *priv;
	LONG err;

	if ((err = DoRawWrite(dio, global_block_current, (void*)buf, HFS_BLOCKSZ, len)) == 0)
		return len;

	ERROR(-1, "error reading from medium");
fail:
	return -1; 
}
